/* ---------------------------------------------------------- */
/*  kome_display_drv.c                                        */
/* ---------------------------------------------------------- */

/*------------------------------------------------------------- */
/** @file
    @brief		V850/CPLD CRTC display driver
    @author		rerofumi
    @since		Jul.23.2007
*/
/*-----------------------------------------------------
 Copyright (C) 2007 rerofumi <rero2@yuumu.org>
 -  Kome wo kame.
 ------------------------------------------------------*/

/*-------------------------------*/
/* include                       */
/*-------------------------------*/

#include "kome_display_drv.h"

#include "mono_ascii.h"

/*-------------------------------*/
/* define                        */
/*-------------------------------*/

#define VRAMADDRESS 0x100000

/*-------------------------------*/
/* struct                        */
/*-------------------------------*/

/*-------------------------------*/
/* global value                  */
/*-------------------------------*/

//! `pPbgobt@
KDisplayPacket  display_packet[KOMEDISPLAY_PACKETSIZE];

//! pPbgNGXg
int  packet_num;
int  packet_ptr_start, packet_ptr_end;

//! CRTC \XCb`
int  crtc_sw;

//! f[^ւ̃|C^e[u
const unsigned char *ascii_ptr_table[128];

/*-------------------------------*/
/* local routine                 */
/*-------------------------------*/

//! pPbg1\
void  put_packet(KDisplayPacket *packet);

//! f[^ւ̃|C^e[upӂ
void  build_ptr_table(void);

/*-------------------------------*/
/* interrupt vector              */
/*-------------------------------*/

/*-------------------------------*/
/* implementation                */
/*-------------------------------*/

/*------------------------------------------------------*/
/* interface                                            */
/*------------------------------------------------------*/

/* --- komeDisplayInitialize() */
/**
 * Display hCȍ
 * |[gݒ
 * 
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayInitialize(void)
{
  int  r;

  packet_num = 0;
  packet_ptr_start = packet_ptr_end = 0;
  crtc_sw = 0;

  build_ptr_table();

  /* --- OoXݒ */
  /* separate bus */
  EXIMC = 0b00000001;
  __asm("nop");
  /* bus size = 8bit */
  BSC = 0x5500;
  /* programable wait = 1cycle */
  DWC0 = 0x0001;
  /* programable address wait = disable */
  AWC = 0xff00;
  /* idle state = disable */
  BCC = 0xaa00;
  /* bus enable */
  PMCCT = 0b00010011;
  PMC9 = 0xffff;
  PMCDH = 0b00111111;
  PMCDL = 0x00ff;
  /* bus control signal */
  PMCCM = 0x02;
  /* --- CRTC signal ݒ */
  PM5 = 0b11111101;
  PMC5 = 0b00000000;
  /* --- CRTC disable */
  P5.1 = 0;
 
  return(DISPLAY_DONE);
}


/* --- komeDisplayFree() */
/**
 * Display hCoI
 * 
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayFree(void)
{
  return(DISPLAY_DONE);
}


/* --- komeDisplayClearPacket() */
/**
 * Display packet ̃NA
 * 
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayClearPacket(void)
{
  packet_num = 0;
  packet_ptr_start = packet_ptr_end = 0;

  return(DISPLAY_DONE);
}



/* --- komeDisplayGetNumPacket() */
/**
 * Display packet ̃NGXg
 *
 * @param num [out] ݂̃NGXgpPbgԂ
 * 
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayGetNumPacket(int *num)
{
  *num = packet_num;

  return(DISPLAY_DONE);
}



/* --- komeDisplaySwCrtc() */
/**
 * CRTC ̕\XCb`
 *
 * @param sw [in] CRTC\Rg[(0=disable, 1=enable)
 * 
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplaySwCrtc(int sw)
{
  crtc_sw = sw;
  /* --- CRTC control */
  if (crtc_sw == 1) {
    /* CRTC enable */
    P5.1 = 1;
  }
  else {
    /* CRTC disable */
    P5.1 = 0;
  }

  return(DISPLAY_DONE);
}



/* --- komeDisplayVSync() */
/**
 * VSync ҂\NGXg
 *
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayVSync(void)
{
	unsigned char  data;

  /* --- CRTC mode check */
  if (crtc_sw == 0) {
    return(DISPLAY_NOT_ENABLE);
  }

	/* --- VBlank҂ */
	while((data = P5.0) == 0) { _asm("nop"); }

  /* --- VBlankȂpPbg\ */
  while(packet_num > 0) {
    /* - VBlank check */
    if ((data = P5.0) == 0) break;
    /* - Packet display */
    put_packet(&(display_packet[packet_ptr_start]));
    packet_ptr_start = (packet_ptr_start + 1) % KOMEDISPLAY_PACKETSIZE;
    packet_num -= 1;
  }
	while((data = P5.0) == 1) { _asm("nop"); }

  return(DISPLAY_DONE);
}


/* --- komeDisplayASync() */
/**
 * ASynchronous \NGXg(CRTC OFFp)
 *
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayASync(void)
{
  while(packet_num > 0) {
    put_packet(&(display_packet[packet_ptr_start]));
    packet_ptr_start = (packet_ptr_start + 1) % KOMEDISPLAY_PACKETSIZE;
    packet_num -= 1;
  }

  return(DISPLAY_DONE);
}



/* --- komeDisplayFill() */
/**
 * `tBNGXg
 *
 * @param x [in] \ʒu x
 * @param y [in] \ʒu y
 * @param w [in] \TCY w
 * @param h [in] \TCY h
 * @param fill [in] tB鐔l(0-255)
 *
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayFill(int x, int y, int w, int h, unsigned char fill)
{
  int  ptr;

  /* --- packet ptr */
  ptr = (packet_ptr_end + 1) % KOMEDISPLAY_PACKETSIZE;
  if (ptr == packet_ptr_start) {
    /*  packet full */
    return(DISPLAY_PACKET_FULL);
  }
  packet_num += 1;
  /* --- packet */
  display_packet[packet_ptr_end].data = 0;
  display_packet[packet_ptr_end].pos_x = x;
  display_packet[packet_ptr_end].pos_y = y;
  display_packet[packet_ptr_end].width = w;
  display_packet[packet_ptr_end].height = h;
  display_packet[packet_ptr_end].fill_data = fill;
  packet_ptr_end = ptr;

  return(DISPLAY_DONE);
}


/* --- komeDisplayClearScreen() */
/**
 * SʃNA
 * 0  fill ē]Ɗ҂pbNɂ
 *
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayClearScreen(void)
{
  komeDisplayClearPacket();
  komeDisplayFill(0, 0, 320, 240, 0);
  if (crtc_sw == 0) {
    komeDisplayASync();
  }
  else {
    komeDisplayVSync();
  }

  return(DISPLAY_DONE);
}


/* --- komeDisplayPattern() */
/**
 * `p^[]NGXg
 *
 * @param x [in] \ʒu x
 * @param y [in] \ʒu y
 * @param w [in] \TCY w
 * @param h [in] \TCY h
 * @param data [in] \p^[f[^|C^
 *
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayPattern(int x, int y, int w, int h, const unsigned char *data)
{
  int  ptr;

  /* --- packet ptr */
  ptr = (packet_ptr_end + 1) % KOMEDISPLAY_PACKETSIZE;
  if (ptr == packet_ptr_start) {
    /*  packet full */
    return(DISPLAY_PACKET_FULL);
  }
  packet_num += 1;
  /* --- packet */
  display_packet[packet_ptr_end].data = (unsigned char *)data;
  display_packet[packet_ptr_end].pos_x = x;
  display_packet[packet_ptr_end].pos_y = y;
  display_packet[packet_ptr_end].width = w;
  display_packet[packet_ptr_end].height = h;
  display_packet[packet_ptr_end].fill_data = 0;
  packet_ptr_end = ptr;

  return(DISPLAY_DONE);
}



/* --- komeDisplayFont() */
/**
 * tHgp^[1\
 *
 * @param x [in] \ʒu x
 * @param y [in] \ʒu y
 * @param code [in] \LN^[
 *
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayFont(int x, int y, int code)
{
  int  ptr;

  /* --- assert */
  if ((code < 0) || (code > 127)) {
    return(DISPLAY_ERROR);
  }

  /* --- packet ptr */
  ptr = (packet_ptr_end + 1) % KOMEDISPLAY_PACKETSIZE;
  if (ptr == packet_ptr_start) {
    /*  packet full */
    return(DISPLAY_PACKET_FULL);
  }
  packet_num += 1;
  /* --- packet */
  display_packet[packet_ptr_end].data = (unsigned char *)ascii_ptr_table[code];
  display_packet[packet_ptr_end].pos_x = x;
  display_packet[packet_ptr_end].pos_y = y;
  display_packet[packet_ptr_end].width = 8;
  display_packet[packet_ptr_end].height = 8;
  display_packet[packet_ptr_end].fill_data = 0;
  packet_ptr_end = ptr;

  return(DISPLAY_DONE);
}


/* --- komeDisplayString() */
/**
 * tHgp^[XgO\
 *
 * @param x [in] \ʒu x
 * @param y [in] \ʒu y
 * @param str [in] \镶|C^
 *
 * @return Ȃ DISPLAY_DONEAȊOŎs
 *
 */
int komeDisplayString(int x, int y, char *str)
{
  int  pos_x, pos_y, safe;
  char *chr;

  chr = str;
  pos_x = x;
  pos_y = y;
  safe = 1000;
  while((*chr != 0x00) && (safe > 0)) {
    komeDisplayFont(pos_x, pos_y, *chr);
    chr += 1;
    safe -= 1;
    pos_x += 8;
    if (pos_x > (320-8)) {
      pos_x = 0;
      pos_y += 8;
      if (pos_y > (240-8)) {
	break;
      }
    }
  }
  
  return(DISPLAY_DONE);
}




/*------------------------------------------------------*/
/* Interrupt                                            */
/*------------------------------------------------------*/
/*------------------------------------------------------*/
/* Local routine                                        */
/*------------------------------------------------------*/

//! pPbg1\
void  put_packet(KDisplayPacket *packet)
{
  int  loop_x, loop_y;
  unsigned char  *buf, pixel;
  unsigned char  *dat;

  dat = packet->data;

  if (dat == 0) {
    pixel = packet->fill_data;
    for(loop_y=0; loop_y<packet->height; loop_y++) {
      buf = packet->pos_x + ((packet->pos_y + loop_y) * 320) + (unsigned char *)VRAMADDRESS;
      for(loop_x=0; loop_x<packet->width; loop_x++) {
        *buf = pixel;
        buf += 1;
      }
    }
  }
  else {
    for(loop_y=0; loop_y<packet->height; loop_y++) {
      buf = packet->pos_x + ((packet->pos_y + loop_y) * 320) + (unsigned char *)VRAMADDRESS;
      for(loop_x=0; loop_x<packet->width; loop_x++) {
        *buf = *dat;
        buf += 1;
        dat += 1;
      }
    }
  }
}


//! f[^ւ̃|C^e[upӂ
void  build_ptr_table(void)
{
  int  i;

  for(i=0; i<128; i++) {
    ascii_ptr_table[i] = 0;
  }
  ascii_ptr_table[0x21] = chr21;
  ascii_ptr_table[0x22] = chr22;
  ascii_ptr_table[0x23] = chr23;
  ascii_ptr_table[0x24] = chr24;
  ascii_ptr_table[0x25] = chr25;
  ascii_ptr_table[0x26] = chr26;
  ascii_ptr_table[0x27] = chr27;
  ascii_ptr_table[0x28] = chr28;
  ascii_ptr_table[0x29] = chr29;
  ascii_ptr_table[0x2a] = chr2a;
  ascii_ptr_table[0x2b] = chr2b;
  ascii_ptr_table[0x2c] = chr2c;
  ascii_ptr_table[0x2d] = chr2d;
  ascii_ptr_table[0x2e] = chr2e;
  ascii_ptr_table[0x2f] = chr2f;
  ascii_ptr_table[0x30] = chr30;
  ascii_ptr_table[0x31] = chr31;
  ascii_ptr_table[0x32] = chr32;
  ascii_ptr_table[0x33] = chr33;
  ascii_ptr_table[0x34] = chr34;
  ascii_ptr_table[0x35] = chr35;
  ascii_ptr_table[0x36] = chr36;
  ascii_ptr_table[0x37] = chr37;
  ascii_ptr_table[0x38] = chr38;
  ascii_ptr_table[0x39] = chr39;
  ascii_ptr_table[0x3a] = chr3a;
  ascii_ptr_table[0x3b] = chr3b;
  ascii_ptr_table[0x3c] = chr3c;
  ascii_ptr_table[0x3d] = chr3d;
  ascii_ptr_table[0x3e] = chr3e;
  ascii_ptr_table[0x3f] = chr3f;
  ascii_ptr_table[0x41] = chr41;
  ascii_ptr_table[0x42] = chr42;
  ascii_ptr_table[0x43] = chr43;
  ascii_ptr_table[0x44] = chr44;
  ascii_ptr_table[0x45] = chr45;
  ascii_ptr_table[0x46] = chr46;
  ascii_ptr_table[0x47] = chr47;
  ascii_ptr_table[0x48] = chr48;
  ascii_ptr_table[0x49] = chr49;
  ascii_ptr_table[0x4a] = chr4a;
  ascii_ptr_table[0x4b] = chr4b;
  ascii_ptr_table[0x4c] = chr4c;
  ascii_ptr_table[0x4d] = chr4d;
  ascii_ptr_table[0x4e] = chr4e;
  ascii_ptr_table[0x4f] = chr4f;
  ascii_ptr_table[0x50] = chr50;
  ascii_ptr_table[0x51] = chr51;
  ascii_ptr_table[0x52] = chr52;
  ascii_ptr_table[0x53] = chr53;
  ascii_ptr_table[0x54] = chr54;
  ascii_ptr_table[0x55] = chr55;
  ascii_ptr_table[0x56] = chr56;
  ascii_ptr_table[0x57] = chr57;
  ascii_ptr_table[0x58] = chr58;
  ascii_ptr_table[0x59] = chr59;
  ascii_ptr_table[0x5a] = chr5a;
  ascii_ptr_table[0x5b] = chr5b;
  ascii_ptr_table[0x5c] = chr5c;
  ascii_ptr_table[0x5d] = chr5d;
  ascii_ptr_table[0x5e] = chr5e;
  ascii_ptr_table[0x5f] = chr5f;
  ascii_ptr_table[0x61] = chr41;
  ascii_ptr_table[0x62] = chr42;
  ascii_ptr_table[0x63] = chr43;
  ascii_ptr_table[0x64] = chr44;
  ascii_ptr_table[0x65] = chr45;
  ascii_ptr_table[0x66] = chr46;
  ascii_ptr_table[0x67] = chr47;
  ascii_ptr_table[0x68] = chr48;
  ascii_ptr_table[0x69] = chr49;
  ascii_ptr_table[0x6a] = chr4a;
  ascii_ptr_table[0x6b] = chr4b;
  ascii_ptr_table[0x6c] = chr4c;
  ascii_ptr_table[0x6d] = chr4d;
  ascii_ptr_table[0x6e] = chr4e;
  ascii_ptr_table[0x6f] = chr4f;
  ascii_ptr_table[0x70] = chr50;
  ascii_ptr_table[0x71] = chr51;
  ascii_ptr_table[0x72] = chr52;
  ascii_ptr_table[0x73] = chr53;
  ascii_ptr_table[0x74] = chr54;
  ascii_ptr_table[0x75] = chr55;
  ascii_ptr_table[0x76] = chr56;
  ascii_ptr_table[0x77] = chr57;
  ascii_ptr_table[0x78] = chr58;
  ascii_ptr_table[0x79] = chr59;
  ascii_ptr_table[0x7a] = chr5a;
}
